==> med_dataloader-0.1.1-py2.py3-none-any.whl: med_dataloader/med_dataloader.py <==
"""Main module."""

import os
import sys
import SimpleITK as sitk
import tensorflow as tf

AUTOTUNE = tf.data.experimental.AUTOTUNE


class DataLoader:
    """[summary]
    """

    def __init__(self,
                 imgA_label,
                 imgB_label,
                 data_dir='./Data',
                 imgs_subdir='Images',
                 cache_mode="prod",
                 extract_only=None,
                 ):
        """[summary]

        Args:
            imgA_label (str): Identifier for class A. It's the name of the
                folder inside :py:attr:`imgs_subdir` that contains images
                labeled as class A.
            imgB_label (str): Identifier for class B. It's the name of the
                folder inside :py:attr:`imgs_subdir` that contains images
                labeled as class B.
            data_dir (str, optional): Path to directory that contains the
                Dataset. This folder **must** contain a subfolder named like
                :py:attr:`imgs_subdir`. Defaults to './Data'.
            imgs_subdir (str, optional): Name (**not** the entire path) of the 
                folder that actually contains :py:attr:`imgA_label` and
                :py:attr:`imgB_label` subfolders. It's a subfolder of Defaults
                to 'Images'.
            cache_mode (str, optional): One between "prod" or "test". "test" is
                used as a debug modality only in case you want to create Cache
                directory at the current folder location. "prod" creates a
                subfolder named "Cache" inside :py:attr:`data_dir`. Defaults to
                "prod".
            extract_only (int, optional): Indicate wheter to partially cache a
                certain amount of elements in the dataset. Please remember that
                if "Cache" folder is already populated, you need to clean this
                folder content to recreate a partial cache file. When it is
                set to None, the entire Dataset is cached. Defaults to None.

        Raises:
            FileNotFoundError: :py:attr:`data_dir` doesn't exists.
            FileNotFoundError: :py:attr:`imgs_subdir` doesn't exists.
            FileNotFoundError: :py:attr:`imgA_label` doesn't exists.
            FileNotFoundError: :py:attr:`imgB_label` doesn't exists.
            ValueError: :py:attr:`cache_mode` is not "prod" nor "test".
        """

        if not os.path.exists(data_dir):
            raise FileNotFoundError(f"{data_dir} does not exist")

        if not os.path.exists(os.path.join(data_dir, imgs_subdir)):
            raise FileNotFoundError(f"{imgs_subdir} does not exist")

        if not os.path.exists(os.path.join(data_dir, imgs_subdir, imgA_label)):
            raise FileNotFoundError(f"{imgA_label} does not exist")

        if not os.path.exists(os.path.join(data_dir, imgs_subdir, imgB_label)):
            raise FileNotFoundError(f"{imgB_label} does not exist")

        self.data_dir = data_dir
        self.imgs_subdir = imgs_subdir
        self.imgA_label = imgA_label
        self.imgB_label = imgB_label

        if cache_mode == "prod":
            self.cache_dir = os.path.join(self.data_dir, "Cache")
        elif cache_mode == "test":
            self.cache_dir = "./Test_Cache"
        else:
            raise ValueError("Cache mode can be only 'prod' or 'test'")
        os.makedirs(self.cache_dir, exist_ok=True)

        self.imgA_paths, self.imgB_paths = self.get_imgs_paths()

        if extract_only is not None:
            self.imgA_paths = self.imgA_paths[:extract_only]
            self.imgB_paths = self.imgB_paths[:extract_only]

    def get_dataset(self,
                    batch_size=32,
                    augmentation=False,
                    crop_size=128):

        ds = tf.data.Dataset.zip((self.get_imgs(img_paths=self.imgA_paths),
                                  self.get_imgs(img_paths=self.imgB_paths)
                                  ))

        ds = ds.map(self.check_dims, num_parallel_calls=AUTOTUNE)

        if augmentation:
            # TODO selectively choose which type of data augmentation to apply
            # ds = ds.map(lambda imgA, imgB: self.random_crop(imgA,
            #                                                imgB,
            #                                                crop_size),
            #            num_parallel_calls=AUTOTUNE)
            ds = ds.map(lambda imgA, imgB: self.resize(imgA,
                                                       imgB,
                                                       crop_size),
                        num_parallel_calls=AUTOTUNE)
            ds = ds.map(self.random_rotate, num_parallel_calls=AUTOTUNE)
            ds = ds.map(self.random_flip, num_parallel_calls=AUTOTUNE)

        ds = ds.batch(batch_size)
        ds = ds.prefetch(buffer_size=AUTOTUNE)

        return ds

    def get_imgs(self, img_paths):
        """Open image files for one class and store it inside cache.

        This function performs all the (usually) slow reading operations that
        is necessary to execute at least the first time. After the first
        execution information are saved inside some cache file inside Cache
        folder (typically created in your Dataset folder, at the same level of
        Images folder). This function detects if cache files are already
        present, and in that case it skips the definition of these files.
        Please take into account that cache files will be as big as your
        Dataset overall size. First execution may result in a considerably
        bigger amount of time.

        Args:
            img_paths (str): Path to single class images.

        Returns:
            tf.Data.Dataset: Tensorflow dataset object containing images of one
                classes converted in Tensor format, without any other
                computations.
        """
        # Get parent folder name from first element in img_paths
        img_label = os.path.basename(os.path.split(img_paths[0])[0])

        cache_file = os.path.join(self.cache_dir, f'{img_label}.cache')
        index_file = f'{cache_file}.index'

        ds = tf.data.Dataset.from_tensor_slices(img_paths)
        ds = ds.map(lambda path: tf.py_function(self.open_img,
                                                [path],
                                                [tf.float32]),
                    num_parallel_calls=AUTOTUNE)
        ds = ds.cache(cache_file)

        if not os.path.exists(index_file):
            self._populate_cache(ds, cache_file, len(img_paths))

        return ds

    def get_imgs_paths(self):
        """Get paths of every single image divided by classes.

        Returns:
            list, list: two list containing the paths of every images for both
                classes. The list is sorted alphabetically, this can be usefull
                when images are named with a progressive number inside a folder
                (e.g.: 001.xxx, 002.xxx, ..., 999.xxx)
        """
        print("Fetching images paths...")
        subset_dir = os.path.join(self.data_dir, self.imgs_subdir)

        subset_dir_imgA = os.path.join(subset_dir, self.imgA_label)
        subset_dir_imgB = os.path.join(subset_dir, self.imgB_label)

        filenames_imgA = os.listdir(subset_dir_imgA)
        filenames_imgB = os.listdir(subset_dir_imgB)

        paths_imgA = [os.path.join(subset_dir_imgA, img)
                      for img in filenames_imgA]
        paths_imgB = [os.path.join(subset_dir_imgB, img)
                      for img in filenames_imgB]

        print("Images paths collected.")

        # Sort paths alphabetically
        paths_imgA.sort()
        paths_imgB.sort()

        return paths_imgA, paths_imgB

    def open_img(self, path):
        """Open an image file and convert it to a tensor.

        Args:
            path (tf.Tensor): Tensor containing the path to the file to be
                opened.

        Returns:
            tf.Tensor: Tensor containing the actual image content.
        """

        path = path.numpy().decode("utf-8")
        image = sitk.GetArrayFromImage(sitk.ReadImage(path))

        # Scale image in 0-1 with a predefined range
        # TODO: remove hardcoding in this function
        image = self.normalise_with_boundaries(image, lb=-1024, ub=3200)
        tensor = tf.convert_to_tensor(image)
        return tf.expand_dims(tensor, axis=-1)

    def _populate_cache(self, ds, cache_file, num_tot):
        print(f'Caching decoded images in {cache_file}...')
        i = 0
        for _ in ds:
            i += 1
            sys.stdout.write('\r')
            sys.stdout.write(f"{i}/{num_tot}")
            sys.stdout.flush()
        print(f'\nCached decoded images in {cache_file}.')

    # -----------------------------------------------------------
    #  Transformations
    # -----------------------------------------------------------
    @staticmethod
    def check_dims(imgA, imgB):
        # TODO: remove hardcoding
        imgA = tf.expand_dims(tf.squeeze(imgA), axis=-1)
        imgA = tf.image.resize_with_pad(imgA, 256, 256)

        imgB = tf.expand_dims(tf.squeeze(imgB), axis=-1)
        imgB = tf.image.resize_with_pad(imgB, 256, 256)

        return imgA, imgB

    @staticmethod
    def normalise_with_boundaries(image, lb, ub):
        """Image normalisation. Normalises image in the range defined by lb and
        ub to fit [0, 1] range."""
        epsilon = 1e-8

        image[image < lb] = lb
        image[image > ub] = ub

        # image = image.astype(np.float32)
        ret = image - lb
        ret /= ((ub - lb) + epsilon)

        return ret

    @staticmethod
    def resize(imgA, imgB, size=256):
        imgA = tf.image.resize(imgA, (size, size))
        imgB = tf.image.resize(imgB, (size, size))

        return imgA, imgB

    @staticmethod
    def random_crop(imgA, imgB, crop_size=256):
        stacked_img = tf.stack([imgA, imgB], axis=0)
        cropped_img = tf.image.random_crop(
            stacked_img, size=[2, crop_size, crop_size, 1])
        cropped_img = tf.split(cropped_img, 2)
        imgA = tf.expand_dims(tf.squeeze(cropped_img[0]), axis=-1)
        imgB = tf.expand_dims(tf.squeeze(cropped_img[1]), axis=-1)
        return imgA, imgB

    @staticmethod
    def random_flip(imgA, imgB):
        if tf.random.uniform(()) > 0.5:
            imgA = tf.image.flip_left_right(imgA)
            imgB = tf.image.flip_left_right(imgB)
        return imgA, imgB

    @staticmethod
    def random_rotate(imgA, imgB):
        rn = tf.random.uniform(shape=(), maxval=4, dtype=tf.int32)
        imgA = tf.image.rot90(imgA, k=rn)
        imgB = tf.image.rot90(imgB, k=rn)
        return imgA, imgB


def generate_dataset(data_loader,
                     percentages,
                     batch_size,
                     input_size
                     ):

    if len(percentages) != 3:
        raise ValueError("Percentages has to be a list of 3 elements")

    if percentages[0] + percentages[1] + percentages[2] != 1.0:
        raise ValueError("Sum of percentages has to be 1")

    num_imgsA = len(data_loader.imgA_paths)
    num_imgsB = len(data_loader.imgB_paths)
    if num_imgsA != num_imgsB:
        raise ValueError("The CBCT and CT subsets have different dimension!")

    train_ends = int(num_imgsA * percentages[0])

    valid_begins = train_ends
    valid_ends = valid_begins + int(num_imgsA * percentages[1])

    complete_ds = data_loader.get_dataset(batch_size=batch_size,
                                          augmentation=True,
                                          crop_size=input_size
                                          )

    complete_ds = complete_ds.unbatch()

    # Train Datasets
    train_ds = complete_ds.take(train_ends)
    train_ds = train_ds.batch(batch_size)

    complete_ds = data_loader.get_dataset(batch_size=batch_size,
                                          augmentation=False)

    complete_ds = complete_ds.unbatch()

    # Validation Datasets
    valid_ds = complete_ds.take(valid_ends)
    valid_ds = valid_ds.skip(valid_begins)
    valid_ds = valid_ds.batch(batch_size)

    # Test Datasets
    test_ds = complete_ds.skip(valid_ends)
    test_ds = test_ds.batch(batch_size)

    return train_ds, valid_ds, test_ds

==> med_dataloader-0.1.6-py2.py3-none-any.whl: med_dataloader/med_dataloader.py <==
"""Main module."""

import os
import sys
import SimpleITK as sitk
import tensorflow as tf

AUTOTUNE = tf.data.experimental.AUTOTUNE


class DataLoader:
    """[summary]"""

    def __init__(
        self,
        imgA_label,
        imgB_label,
        data_dir="./Data",
        imgs_subdir="Images",
        cache_mode="prod",
        extract_only=None,
    ):
        """[summary]

        Args:
            imgA_label (str): Identifier for class A. It's the name of the
                folder inside :py:attr:`imgs_subdir` that contains images
                labeled as class A.
            imgB_label (str): Identifier for class B. It's the name of the
                folder inside :py:attr:`imgs_subdir` that contains images
                labeled as class B.
            data_dir (str, optional): Path to directory that contains the
                Dataset. This folder **must** contain a subfolder named like
                :py:attr:`imgs_subdir`. Defaults to './Data'.
            imgs_subdir (str, optional): Name (**not** the entire path) of the
                folder that actually contains :py:attr:`imgA_label` and
                :py:attr:`imgB_label` subfolders. It's a subfolder of Defaults
                to 'Images'.
            cache_mode (str, optional): One between "prod" or "test". "test" is
                used as a debug modality only in case you want to create Cache
                directory at the current folder location. "prod" creates a
                subfolder named "Cache" inside :py:attr:`data_dir`. Defaults to
                "prod".
            extract_only (int, optional): Indicate wheter to partially cache a
                certain amount of elements in the dataset. Please remember that
                if "Cache" folder is already populated, you need to clean this
                folder content to recreate a partial cache file. When it is
                set to None, the entire Dataset is cached. Defaults to None.

        Raises:
            FileNotFoundError: :py:attr:`data_dir` doesn't exists.
            FileNotFoundError: :py:attr:`imgs_subdir` doesn't exists.
            FileNotFoundError: :py:attr:`imgA_label` doesn't exists.
            FileNotFoundError: :py:attr:`imgB_label` doesn't exists.
            ValueError: :py:attr:`cache_mode` is not "prod" nor "test".
        """

        if not os.path.exists(data_dir):
            raise FileNotFoundError(f"{data_dir} does not exist")

        if not os.path.exists(os.path.join(data_dir, imgs_subdir)):
            raise FileNotFoundError(f"{imgs_subdir} does not exist")

        if not os.path.exists(os.path.join(data_dir, imgs_subdir, imgA_label)):
            raise FileNotFoundError(f"{imgA_label} does not exist")

        if not os.path.exists(os.path.join(data_dir, imgs_subdir, imgB_label)):
            raise FileNotFoundError(f"{imgB_label} does not exist")

        self.data_dir = data_dir
        self.imgs_subdir = imgs_subdir
        self.imgA_label = imgA_label
        self.imgB_label = imgB_label

        if cache_mode == "prod":
            self.cache_dir = os.path.join(self.data_dir, "Cache")
        elif cache_mode == "test":
            self.cache_dir = "./Test_Cache"
        else:
            raise ValueError("Cache mode can be only 'prod' or 'test'")
        os.makedirs(self.cache_dir, exist_ok=True)

        self.imgA_paths, self.imgB_paths = self.get_imgs_paths()

        if extract_only is not None:
            self.imgA_paths = self.imgA_paths[:extract_only]
            self.imgB_paths = self.imgB_paths[:extract_only]

        self.is_3D = self.is_3D_data(self.imgA_paths[0])
        self.imgA_type = self.check_type(self.imgA_paths[0])
        self.imgB_type = self.check_type(self.imgB_paths[0])

    def get_dataset(self,
                    input_size,
                    batch_size=32,
                    norm_boundsA=None,
                    norm_boundsB=None,
                    augmentation=False,
                    random_crop_size=None,
                    random_rotate=False,
                    random_flip=False):

        ds = tf.data.Dataset.zip((self.get_imgs(img_paths=self.imgA_paths,
                                                img_type=self.imgA_type,
                                                norm_bounds=norm_boundsA),
                                  self.get_imgs(img_paths=self.imgB_paths,
                                                img_type=self.imgB_type,
                                                norm_bounds=norm_boundsB)
                                  ))

        ds = ds.map(lambda imgA, imgB: self.check_dims(imgA,
                                                       imgB,
                                                       input_size),
                    num_parallel_calls=AUTOTUNE)

        if augmentation:
            if random_crop_size:
                ds = ds.map(
                    lambda imgA, imgB: self.random_crop(imgA,
                                                        imgB,
                                                        random_crop_size),
                    num_parallel_calls=AUTOTUNE)
            if random_rotate:
                ds = ds.map(self.random_rotate, num_parallel_calls=AUTOTUNE)
            if random_flip:
                ds = ds.map(self.random_flip, num_parallel_calls=AUTOTUNE)

        ds = ds.batch(batch_size)
        ds = ds.prefetch(buffer_size=AUTOTUNE)

        return ds

    def get_imgs(self, img_paths, img_type, norm_bounds=None):
        """Open image files for one class and store it inside cache.

        This function performs all the (usually) slow reading operations that
        is necessary to execute at least the first time. After the first
        execution information are saved inside some cache file inside Cache
        folder (typically created in your Dataset folder, at the same level of
        Images folder). This function detects if cache files are already
        present, and in that case it skips the definition of these files.
        Please take into account that cache files will be as big as your
        Dataset overall size. First execution may result in a considerably
        bigger amount of time.

        Args:
            img_paths (str): Path to single class images.

        Returns:
            tf.Data.Dataset: Tensorflow dataset object containing images of one
                classes converted in Tensor format, without any other
                computations.
        """
        # Get parent folder name from first element in img_paths
        img_label = os.path.basename(os.path.split(img_paths[0])[0])

        cache_file = os.path.join(self.cache_dir, f"{img_label}.cache")
        index_file = f"{cache_file}.index"

        ds = tf.data.Dataset.from_tensor_slices(img_paths)
        ds = ds.map(lambda path: tf.py_function(self.open_img,
                                                [path],
                                                [img_type],
                                                ),
                    num_parallel_calls=AUTOTUNE)

        if norm_bounds is not None:
            ds = ds.map(lambda img: self.norm_with_bounds(img,
                                                          norm_bounds),
                        num_parallel_calls=AUTOTUNE
                        )

        if self.is_3D:
            ds = ds.unbatch()

        ds = ds.cache(cache_file)

        if not os.path.exists(index_file):
            self._populate_cache(ds, cache_file, len(img_paths))

        return ds

    def get_imgs_paths(self):
        """Get paths of every single image divided by classes.

        Returns:
            list, list: two list containing the paths of every images for both
                classes. The list is sorted alphabetically, this can be usefull
                when images are named with a progressive number inside a folder
                (e.g.: 001.xxx, 002.xxx, ..., 999.xxx)
        """
        # print("Fetching images paths...")
        subset_dir = os.path.join(self.data_dir, self.imgs_subdir)

        subset_dir_imgA = os.path.join(subset_dir, self.imgA_label)
        subset_dir_imgB = os.path.join(subset_dir, self.imgB_label)

        filenames_imgA = os.listdir(subset_dir_imgA)
        filenames_imgB = os.listdir(subset_dir_imgB)

        paths_imgA = [os.path.join(subset_dir_imgA, img)
                      for img in filenames_imgA]
        paths_imgB = [os.path.join(subset_dir_imgB, img)
                      for img in filenames_imgB]

        # print("Images paths collected.")

        # Sort paths alphabetically
        paths_imgA.sort()
        paths_imgB.sort()

        return paths_imgA, paths_imgB

    def open_img(self, path):
        """Open an image file and convert it to a tensor.

        Args:
            path (tf.Tensor): Tensor containing the path to the file to be
                opened.

        Returns:
            tf.Tensor: Tensor containing the actual image content.
        """

        path = path.numpy().decode("utf-8")
        image = sitk.GetArrayFromImage(sitk.ReadImage(path))

        tensor = tf.convert_to_tensor(image)

        return tf.expand_dims(tensor, axis=-1)

    def _populate_cache(self, ds, cache_file, num_tot):
        print(f"Caching decoded images in {cache_file}...")
        i = 0
        for _ in ds:
            i += 1
            sys.stdout.write("\r")
            sys.stdout.write(f"{i}/{num_tot}")
            sys.stdout.flush()
        print(f"\nCached decoded images in {cache_file}.")

    @staticmethod
    def is_3D_data(path):
        image = sitk.GetArrayFromImage(sitk.ReadImage(path))

        if len(image.shape) == 3:
            return True
        elif len(image.shape) == 2:
            return False
        else:
            raise ValueError("Work only with 2D or 3D files.")

        return True

    @staticmethod
    def check_type(path):
        image = sitk.GetArrayFromImage(sitk.ReadImage(path))

        dict_type = {"int8": tf.int8,
                     "int16": tf.int16,
                     "int32": tf.int32,
                     "int64": tf.int64,
                     "uint8": tf.uint8,
                     "uint16": tf.uint16,
                     "uint32": tf.uint32,
                     "uint64": tf.uint64,
                     "float32": tf.float32,
                     "float64": tf.float64,
                     }

        img_type = image.dtype.name

        return dict_type[img_type]

    @ staticmethod
    def check_dims(imgA, imgB, size):
        imgA = tf.expand_dims(tf.squeeze(imgA), axis=-1)
        imgA = tf.image.resize_with_pad(imgA, size, size)

        imgB = tf.expand_dims(tf.squeeze(imgB), axis=-1)
        imgB = tf.image.resize_with_pad(imgB, size, size)

        return imgA, imgB

    # -----------------------------------------------------------
    #  Transformations
    # -----------------------------------------------------------
    @ staticmethod
    def norm_with_bounds(image, bounds):
        """Image normalisation. Normalises image in the range defined by lb and
        ub to fit [0, 1] range."""
        epsilon = 1e-8
        lb = bounds[0]
        ub = bounds[1]

        tf.where(image < lb, lb, image)
        tf.where(image > ub, ub, image)

        image = image - lb
        image /= (ub - lb) + epsilon

        return image

    @ staticmethod
    def random_crop(imgA, imgB, crop_size=256):
        stacked_img = tf.stack([imgA, imgB], axis=0)
        cropped_img = tf.image.random_crop(
            stacked_img, size=[2, crop_size, crop_size, 1]
        )
        cropped_img = tf.split(cropped_img, 2)
        imgA = tf.expand_dims(tf.squeeze(cropped_img[0]), axis=-1)
        imgB = tf.expand_dims(tf.squeeze(cropped_img[1]), axis=-1)
        return imgA, imgB

    @ staticmethod
    def random_flip(imgA, imgB):
        if tf.random.uniform(()) > 0.5:
            imgA = tf.image.flip_left_right(imgA)
            imgB = tf.image.flip_left_right(imgB)
        return imgA, imgB

    @ staticmethod
    def random_rotate(imgA, imgB):
        rn = tf.random.uniform(shape=(), maxval=4, dtype=tf.int32)
        imgA = tf.image.rot90(imgA, k=rn)
        imgB = tf.image.rot90(imgB, k=rn)
        return imgA, imgB


def generate_dataset(data_loader,
                     input_size,
                     percentages,
                     batch_size,
                     norm_boundsA=None,
                     norm_boundsB=None,
                     train_augmentation=True,
                     random_crop_size=None,
                     random_rotate=True,
                     random_flip=True,
                     ):

    if len(percentages) != 3:
        raise ValueError("Percentages has to be a list of 3 elements")

    if percentages[0] + percentages[1] + percentages[2] != 1.0:
        raise ValueError("Sum of percentages has to be 1")

    num_imgsA = len(data_loader.imgA_paths)
    num_imgsB = len(data_loader.imgB_paths)
    if num_imgsA != num_imgsB:
        raise ValueError("The CBCT and CT subsets have different dimension!")

    complete_ds = data_loader.get_dataset(batch_size=batch_size,
                                          norm_boundsA=norm_boundsA,
                                          norm_boundsB=norm_boundsB,
                                          input_size=input_size,
                                          augmentation=train_augmentation,
                                          random_crop_size=random_crop_size,
                                          random_rotate=random_rotate,
                                          random_flip=random_flip)

    complete_ds = complete_ds.unbatch()

    # Compute length of dataset
    num_imgs = 0
    for img in complete_ds:
        num_imgs += 1

    train_ends = int(num_imgs * percentages[0])

    valid_begins = train_ends
    valid_ends = valid_begins + int(num_imgs * percentages[1])

    # Train Datasets
    train_ds = complete_ds.take(train_ends)
    train_ds = train_ds.batch(batch_size)

    # Same as before, but without augmentation since now we want to obtain
    # validation and test set
    complete_ds = data_loader.get_dataset(batch_size=batch_size,
                                          norm_boundsA=norm_boundsA,
                                          norm_boundsB=norm_boundsB,
                                          input_size=input_size,
                                          augmentation=False)

    complete_ds = complete_ds.unbatch()

    # Validation Datasets
    valid_ds = complete_ds.take(valid_ends)
    valid_ds = valid_ds.skip(valid_begins)
    valid_ds = valid_ds.batch(batch_size)

    # Test Datasets
    test_ds = complete_ds.skip(valid_ends)
    test_ds = test_ds.batch(batch_size)

    return train_ds, valid_ds, test_ds
